{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "source": [
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/BrainLab-Rectangle-LightBG-Color@5x.png\" width=\"600\"/>"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Welcome to BrainLab's hands-on training sessions! The purpose is to illustrate the concepts behind the demos in the course in a way that allows you to explore different functions, arguments and examples as you master the material."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Recall: Blockchain (Non-Distributed)\n",
    "\n",
    "\n",
    "In the previous hands-on session, we learned what it takes to tie blocks together in a blockchain, and defined `Block` and `Blockchain` objects.\n",
    "\n",
    "We'll start by redefining the objects with a few minor additions:\n",
    "\n",
    "- The `Block` object will include an `__eq__(self, other)` method in order to be able to determine whether its contents are similar to another block's (we'll compare its `number`, `nonce`, `data`, `prev_hash` and `hash` fields).\n",
    "- The `Blockchain` object will also include an `__eq__(self, other)` in order to be able to determine whether its blocks are similar to another blockchain's (we'll individually compare its blocks). \n",
    "- The `Blockchain` object will include a `copy(self)` method in order for us to be able to easily replicate it.\n",
    "\n",
    "These will be essential in order for us to easily copy a blockchain and see what happens when every user on the network has her own copy!\n",
    "\n",
    "Recall that the overarching premise of the blockchain is that it has to be **immutable**. If one party modifies her copy, nefariously or otherwise, then the rest of the network has to be able to spot it."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 68,
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "import hashlib\n",
    "\n",
    "\n",
    "class Block:\n",
    "    \"\"\"Block class to package all relevant data and helper methods.\"\"\"\n",
    "    \n",
    "    def __init__(self, blockchain, is_first_block=False):\n",
    "        \"\"\"Initiate the block under the assumption that it must either be\n",
    "        (1) The first block in the chain, in which case its data will be prepopulated or\n",
    "        (2) Any other block, in which case we'll be able to derive some fields from the\n",
    "            last block on the chain.\n",
    "        \"\"\"\n",
    "        \n",
    "        if is_first_block:\n",
    "            # As the first block, its number will naturally be \"1\".\n",
    "            self.number = \"1\"\n",
    "            \n",
    "            # Its previous block's hash will simply consist of zeros (64 zeros, since\n",
    "            # that's the number of characters in a sha-256 hash), data field will be blank\n",
    "            # and a nonce of \"11316\" will be used to create a hash that meets the required\n",
    "            # level of difficulty.\n",
    "            self.prev_hash = \"0\" * 64\n",
    "            self.nonce = \"11316\"\n",
    "            self.set_data(\"\")\n",
    "\n",
    "        else:\n",
    "            # Since this is not the first block on the chain, we will need to derive some\n",
    "            # information from the last block on the current blockchain.\n",
    "            last_block = blockchain.get_last_block()\n",
    "            \n",
    "            # Increment the last block's number to get its own.\n",
    "            self.number = int(last_block.number) + 1\n",
    "            \n",
    "            # Base its last block's hash on the previous block's hash.\n",
    "            self.prev_hash = last_block.hash\n",
    "            \n",
    "            # Required fields, which will be populated via separate methods later.\n",
    "            self.nonce = None\n",
    "            self.data = None\n",
    "            self.hash = None\n",
    "        \n",
    "    \n",
    "    def set_data(self, data):\n",
    "        \"\"\"Update the data of the block as well as its own hash.\"\"\"\n",
    "        \n",
    "        self.data = data\n",
    "        self.set_hash()\n",
    "\n",
    "    \n",
    "    def set_prev_hash(self, prev_hash):\n",
    "        \"\"\"Update the previous hash field of the block as well as its own hash.\"\"\"\n",
    "        \n",
    "        self.prev_hash = prev_hash\n",
    "        self.set_hash()\n",
    "        \n",
    "\n",
    "    def set_hash(self):\n",
    "        \"\"\"Create a simple block's hash by concatenating all required fields.\"\"\"\n",
    "        \n",
    "        block_fields =  str(self.number) + \\\n",
    "                        str(self.nonce)  + \\\n",
    "                        str(self.data)   + \\\n",
    "                        str(self.prev_hash)\n",
    "        \n",
    "        # Use Python's native SHA-256 function.\n",
    "        self.hash = hashlib.sha256(block_fields.encode(\"utf-8\")).hexdigest()\n",
    "        \n",
    "        \n",
    "    def has_valid_hash(self):\n",
    "        \"\"\"Determine the validity of a hash, based on the number of zeros it starts with.\"\"\"\n",
    "        \n",
    "        difficulty = 4\n",
    "        return self.hash.startswith(\"0\" * difficulty)\n",
    "    \n",
    "        \n",
    "    def mine(self):\n",
    "        \"\"\"Obtain a nonce for the block that leads to a valid hash, considering its other fields.\"\"\"\n",
    "\n",
    "        # Start with a guess of \"1\", and increment it until we obtain a valid hash. At that point,\n",
    "        # store the last nonce value.\n",
    "        self.nonce = 1\n",
    "        \n",
    "        while self.nonce:\n",
    "            self.set_hash()\n",
    "\n",
    "            if self.has_valid_hash():\n",
    "                break\n",
    "\n",
    "            self.nonce += 1\n",
    "        \n",
    "        \n",
    "    def __eq__(self, other):\n",
    "        \"\"\"Determine whether this block's fields are similar to the other's.\"\"\"\n",
    "        \n",
    "        return self.__dict__ == other.__dict__\n",
    "        \n",
    "    \n",
    "    def __str__(self):\n",
    "        \"\"\"Use to print the contents of the block in a helpful manner.\"\"\"\n",
    "        \n",
    "        output = \"Block |\"\n",
    "        \n",
    "        output += \"\\tNumber\\t\\t=>\\t\"   + str(self.number)     + \"\\n\"\n",
    "        output += \"\\tNonce\\t\\t=>\\t\"    + str(self.nonce)      + \"\\n\"\n",
    "        output += \"\\tData\\t\\t=>\\t\"     + str(self.data)       + \"\\n\"\n",
    "        output += \"\\tPrev. Hash\\t=>\\t\" + str(self.prev_hash)  + \"\\n\"\n",
    "        output += \"\\tHash\\t\\t=>\\t\"     + str(self.hash)\n",
    "        \n",
    "        return output"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 69,
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Blockchain:\n",
    "    \"\"\"Blockchain class to package all relevant blocks and helper methods.\"\"\"\n",
    "    \n",
    "    def __init__(self):\n",
    "        \"\"\"Create an empty list to store our chain of blocks, or blockchain!\"\"\"\n",
    "        \n",
    "        self.blocks = []\n",
    "        \n",
    "        \n",
    "    def append_block(self, block):\n",
    "        \"\"\"Append a block object (defined above) to our chain of blocks.\"\"\"\n",
    "        \n",
    "        self.blocks.append(block)\n",
    "    \n",
    "        \n",
    "    def get_last_block(self):\n",
    "        \"\"\"Return the most recent block in the chain.\"\"\"\n",
    "        \n",
    "        return self.blocks[-1]\n",
    "    \n",
    "    \n",
    "    def validate(self):\n",
    "        \"\"\"Ensure that our blockchain is valid by:\n",
    "        (1) Verifying that every block has a valid hash. If an older block changes its\n",
    "            data and does not mine for a new nonce, then this check will fail.\n",
    "        (2) Verifying that every block's previous hash field links to the correct\n",
    "            previous block. If an older block is tampered with and its hash changes without\n",
    "            the knowledge of the next block in the chain, then this check will fail.\n",
    "        \"\"\"\n",
    "        \n",
    "        # In a chain of blocks A -> B -> C, this variable is used to compare C's previous\n",
    "        # hash field against B's own hash field.\n",
    "        expected_prev_hash = None\n",
    "        \n",
    "        # Iterate through the blocks, starting with the most recent one.\n",
    "        for block in reversed(self.blocks):\n",
    "            \n",
    "            # Check (1) above. Specify and output the number of the guilty block!\n",
    "            if not block.has_valid_hash():\n",
    "                print(\"Invalid blockchain due to invalid hash.\")\n",
    "                print(\"The guilty block is: #%s.\" % block.number)\n",
    "                break\n",
    "            \n",
    "            # Check (2) above. Specify and output the number of the guilty block!\n",
    "            if expected_prev_hash and (block.hash != expected_prev_hash):\n",
    "                print(\"Invalid blockchain due to mismatched prev_hash.\")\n",
    "                print(\"The guilty block is: #%s.\" % block.number)\n",
    "                break\n",
    "            \n",
    "            # Store the current block's previous hash as stated above.\n",
    "            expected_prev_hash = block.prev_hash\n",
    "        \n",
    "        # Conclude that our blockchain is valid if we didn't find any faulty blocks\n",
    "        # and, therefore, did not have to break out of the checking process.\n",
    "        else:       \n",
    "            print(\"Valid blockchain!\")\n",
    "        \n",
    "    \n",
    "    def copy(self):\n",
    "        \"\"\"Creates a full copy of the blockchain (including its blocks).\n",
    "        https://docs.python.org/3/library/copy.html#module-copy\n",
    "        \"\"\"\n",
    "        \n",
    "        return copy.deepcopy(self.blocks)\n",
    "    \n",
    "    \n",
    "    def __eq__(self, other): \n",
    "        \"\"\"Determine whether this blockchain's blocks are similar to the other's.\"\"\"\n",
    "        \n",
    "        for count, block in enumerate(self.blocks):\n",
    "            if block != other.blocks[count]:\n",
    "                return False\n",
    "            \n",
    "        return True\n",
    "    \n",
    "    \n",
    "    def __str__(self):\n",
    "        \"\"\"Use to print the contents of the blockchain in a helpful manner.\"\"\"\n",
    "        \n",
    "        output = \"\\n\\n\"\n",
    "        \n",
    "        for block in self.blocks:\n",
    "            output += block.__str__() + \"\\n\\n\\n\"\n",
    "        \n",
    "        return output"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "# Distributed Blockchain\n",
    "\n",
    "We say that the blockchain is secure and immutable precisely because it is distributed and prepared to have **tens of thousands of copies around the world** (for instance, Bitcoin's blockchain).\n",
    "\n",
    "In a network with this many copies of the blockchain, even if a single instance is compromised or tampered with, the data will endure via the other copies of the ledger.\n",
    "\n",
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/brainlab-demo-distributed-blockchain.png\" width=\"600\" />\n",
    "\n",
    "We'll now create a `DistributedBlockchain` object that will wrap our `Blockchain` object above with information about its owner, or `peer`. Note that this doesn't mean that the `peer` will *own* the blockchain — since nobody does — but that she may have a *copy* of the full blockchain running on her own infrastructure."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 66,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class DistributedBlockchain:\n",
    "    \"\"\"Distributed blockchain class to package its peer, relevant blocks and helper methods.\"\"\"\n",
    "    \n",
    "    def __init__(self, peer, blockchain=None):\n",
    "        \"\"\"Keep track of the peer (owner of this instance of the blockchain) as well as on\n",
    "        the blockchain object above with the relevant blocks.\n",
    "        \"\"\"\n",
    "        \n",
    "        self.peer = peer\n",
    "        \n",
    "        # Store a blockchain if it's provided via the constructor.\n",
    "        if blockchain:\n",
    "            self.blockchain = blockchain\n",
    "        \n",
    "        # If a blockchain is not provided to start it, start a fresh instance with a\n",
    "        # prepopulated initial block.\n",
    "        else:\n",
    "            self.blockchain = Blockchain()\n",
    "\n",
    "            first_block = Block(self.blockchain, is_first_block=True)\n",
    "            self.append_block(first_block)\n",
    "        \n",
    "        \n",
    "    def append_block(self, block):\n",
    "        \"\"\"Append a block object (defined above) to our chain of blocks.\"\"\"\n",
    "        \n",
    "        self.blockchain.append_block(block)\n",
    "                \n",
    "        \n",
    "    def __eq__(self, other):\n",
    "        \"\"\"Determine whether this blockchain is similar to the other's.\"\"\"\n",
    "        \n",
    "        return self.blockchain == other.blockchain\n",
    "    \n",
    "    \n",
    "    def __str__(self):\n",
    "        \"\"\"Use to print the contents of the distributed blockchain in a helpful manner.\"\"\"\n",
    "        \n",
    "        output = \"Blockchain | \" + str(self.peer) + \"\\n\"\n",
    "        output += self.blockchain.__str__()\n",
    "        \n",
    "        return output"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Examples\n",
    "\n",
    "We'll now dive into some examples, where three participants known as `vivian`, `joseph` and `george` will have their own copies of a blockchain with three blocks.\n",
    "\n",
    "Then, we'll go into the different scenarios that will show what happens when one party tries to deviate from the rest."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 87,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Blockchain | Vivian\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tTuesday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Define a sample blockchain under Vivian's account.\n",
    "vivian = DistributedBlockchain('Vivian')\n",
    "\n",
    "# Create, mine and append a second block.\n",
    "block2 = Block(vivian.blockchain)\n",
    "block2.set_data('Tuesday')\n",
    "block2.mine()\n",
    "vivian.append_block(block2)\n",
    "\n",
    "# Create, mine and append a third block.\n",
    "block3 = Block(vivian.blockchain)\n",
    "block3.set_data('Wednesday.')\n",
    "block3.mine()\n",
    "vivian.append_block(block3)\n",
    "\n",
    "# Let's see how our copy of the blockchain is doing thus far.\n",
    "print(vivian)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "We'll now incorporate `Joseph` and `George` into our network. Say, for instance, that they've just learned about our incredible blockchain and are now looking to house it fully on their own machines in order to become miners.\n",
    "\n",
    "The first step for them will be to make full copies of the blockchain, as shown below!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 88,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Blockchain | Joseph\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tTuesday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n",
      "Blockchain | George\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tTuesday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Instantiate copies of Vivian's instance of the blockchain for Joseph and George.\n",
    "joseph = DistributedBlockchain('Joseph', vivian.copy())\n",
    "george = DistributedBlockchain('George', vivian.copy())\n",
    "\n",
    "# We'll now print Joseph's and George's instances, in order to verify that we\n",
    "# actually got full copies of Vivian's instance of the blockchain.\n",
    "print(joseph)\n",
    "print(george)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 89,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "True\n",
      "True\n",
      "True\n"
     ]
    }
   ],
   "source": [
    "# We can also use our equality operators to verify that every block is similar.\n",
    "print(joseph.blockchain == vivian.blockchain)\n",
    "print(george.blockchain == vivian.blockchain)\n",
    "print(joseph.blockchain == george.blockchain)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "It's important to know that we're dealing with **three distributed copies** of the blockchain, not a single blockchain that any one of these parties may change.\n",
    "\n",
    "The latter scenario would be extremely dangerous, since it'd mean that one party can modify the common ledger that we're talking about. Precisely, what's interesting about the blockchain is that we can trust parties we don't know to have copies of the same ledger.\n",
    "\n",
    "What happens, then, if Joseph decides to modify the data in block #2? How will that affect what Vivian and George know about the same blockchain?\n",
    "\n",
    "Let's try it out!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 90,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Blockchain | Joseph\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tFriday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t86191e900654a45b64a238e94d42562d3211f7f963d2e84ebb88e5aae370aeb1\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Let's change the data of block #2 to say \"Friday\" instead of \"Tuesday\"\n",
    "# in Joseph's blockchain. (Note that, since the list is implemented in Python,\n",
    "# whose lists start with 0, the second block will be accessed via the 1 index).\n",
    "joseph_block2 = joseph.blockchain.blocks[1]\n",
    "joseph_block2.set_data('Friday')\n",
    "\n",
    "# How does this impact Joseph's blockchain?\n",
    "print(joseph)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Clearly, we can compare Joseph's blockchain against Vivian's and George's above to see that it has changed. Its second block is no longer similar to the other blockchain's, as its hash is not even valid!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 91,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Blockchain | Vivian\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tTuesday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n",
      "Blockchain | George\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t24765\n",
      "\tData\t\t=>\tTuesday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t67551\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t0000b251bd673dd811aa01145d202b1a6a695a57e070cb5e83f542a39c219118\n",
      "\tHash\t\t=>\t000017c9d2fe10444fd5cc2d679e89162932b071511474a148072c90546ce025\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Visualize Vivian's and George's blockchains to verify that Joseph's differs.\n",
    "print(vivian)\n",
    "print(george)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 92,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "True\n",
      "False\n",
      "False\n"
     ]
    }
   ],
   "source": [
    "# In fact, while Vivian's and George's remain similar, they don't\n",
    "# match Joseph's!\n",
    "print(vivian.blockchain == george.blockchain)\n",
    "print(joseph.blockchain == vivian.blockchain)\n",
    "print(joseph.blockchain == george.blockchain)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Now, as we saw in the previous hands-on sessions, it's important to mention that Joseph could re-mine his second block in order to obtain a valid hash, and then re-mine the third in order to incorporate the second block's hash in its `prev_hash` field.\n",
    "\n",
    "Let's see how that could *fix* Joseph's blockchain, and how it would compare to Vivian's and George's."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 97,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Invalid blockchain due to mismatched prev_hash.\n",
      "The guilty block is: #2.\n"
     ]
    }
   ],
   "source": [
    "# Re-mine Joseph's second block and try to validate his full copy of the blockchain.\n",
    "joseph_block2.mine()\n",
    "\n",
    "joseph.blockchain.validate()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 100,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Valid blockchain!\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t1\n",
      "\tNonce\t\t=>\t11316\n",
      "\tData\t\t=>\t\n",
      "\tPrev. Hash\t=>\t0000000000000000000000000000000000000000000000000000000000000000\n",
      "\tHash\t\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t2\n",
      "\tNonce\t\t=>\t117769\n",
      "\tData\t\t=>\tFriday\n",
      "\tPrev. Hash\t=>\t000015783b764259d382017d91a36d206d0600e2cbb3567748f46a33fe9297cf\n",
      "\tHash\t\t=>\t000047f3537304790973f92a56b10aeb73b1192e4e91badbcca9fc8a98ddf59c\n",
      "\n",
      "\n",
      "Block |\tNumber\t\t=>\t3\n",
      "\tNonce\t\t=>\t178815\n",
      "\tData\t\t=>\tWednesday.\n",
      "\tPrev. Hash\t=>\t000047f3537304790973f92a56b10aeb73b1192e4e91badbcca9fc8a98ddf59c\n",
      "\tHash\t\t=>\t000007049a7f466d949996ec699edb8c1cf80bd9a17c8bf914c20afafc22e729\n",
      "\n",
      "\n",
      "\n"
     ]
    }
   ],
   "source": [
    "# Indeed, we have to set block #3's prev_hash field to match block #2's hash.\n",
    "joseph_block3 = joseph.blockchain.blocks[2]\n",
    "joseph_block3.prev_hash = joseph_block2.hash\n",
    "joseph_block3.mine()\n",
    "\n",
    "# Is Joseph's blockchain valid now?\n",
    "joseph.blockchain.validate()\n",
    "\n",
    "print(joseph.blockchain)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Joseph's blockchain is valid now, so we could think that he got away with changing the value of the data field in block #2 to \"Friday\", thereby violating the principle of immutability that we've emphasized the blockchain is all about!\n",
    "\n",
    "However, that's not quite the case.\n",
    "\n",
    "Even if Joseph's blockchain does not contain any invalid hashes or broken links, it's fairly useless insofar as it does not match Vivian's or George's versions."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 101,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "True\n",
      "False\n",
      "False\n"
     ]
    }
   ],
   "source": [
    "# Compare Joseph's blockchain to Vivian's and George's.\n",
    "print(joseph.blockchain == vivian.blockchain)\n",
    "print(joseph.blockchain == george.blockchain)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Joseph has effectively **forked** his blockchain away from Vivian's and George's.\n",
    "\n",
    "In order to effectively modify a block in his version of the blockchain _and_ remain complicit with Vivian and George, he would have to modify Vivian's and George's versions as well.\n",
    "\n",
    "Now, if we apply these concepts to a popular blockchain such as Bitcoin's, then Joseph would not only have had to _hack_ into Vivian's and George's, but also the other tens of thousands of nodes in the network... an exceedingly difficult feat in terms of computing resources and cybersecurity."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "# Congratulations!\n",
    "\n",
    "We've now gone through the hands-on training sessions of BrainLab's \"Blockchain for Absolute Beginners\" course.\n",
    "\n",
    "We started by exploring the concept and importance of hashing insofar as it allowed us to create digital signatures of our data and verify their immutability.\n",
    "\n",
    "Then, we created blocks that encompassed block numbers, nonces, data and hashes into single items (which we carried onto the `Block` object above).\n",
    "\n",
    "Blockchains, represented by the `Blockchain` object, taught us to tie blocks together via their `prev_hash` fields, which also served to ensure their immutability.\n",
    "\n",
    "Lastly, we saw how a distributed and replicated blockchain lets an unbounded number of parties come together in a network to maintain copies of the same ledger. And, by that logic, ensure that an individual who modifies the data in a single instance does not affect the data that everyone else has."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "<img src=\"https://ibleducation.com/wp-content/uploads/2018/12/BrainLab-Rectangle-LightBG-Color@5x.png\" width=\"600\"/>"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.5.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
